// ServoDecodeTest

#define DEBUGASERIAL	1

// Times for 57600 baud
#if F_CPU == 20000000L   // 20MHz clock 
  // Sinan: Not tested                                                     
  #define TICKS2COUNT         348  // Ticks between two bits.
  #define TICKS2WAITONE       348  // Wait one bit period.
  #define TICKS2WAITONE_HALF  520    // Wait one and a half bit period.
#elif F_CPU == 16000000L  // 16MHz clock                                                  
  #define TICKS2COUNT         278  // Ticks between two bits.
  #define TICKS2WAITONE       278  // Wait one bit period.
  #define TICKS2WAITONE_HALF  416    // Wait one and a half bit period.
#elif F_CPU == 8000000L   // 8MHz clock
  // Assumes a 8MHz clock                                                   
  #define TICKS2COUNT         139  // Ticks between two bits.
  #define TICKS2WAITONE       139  // Wait one bit period.
  #define TICKS2WAITONE_HALF  208    // Wait one and a half bit period.
#else
  #error Unsupported clock speed
#endif

#if F_CPU == 20000000L     // 20MHz clock                                                  
  #define INTERRUPT_EXEC_CYCL   112       // Cycles to execute interrupt routines from interrupt.
  #define INTERRUPT_EARLY_BIAS  40       // Cycles to allow of other interrupts.

#elif F_CPU == 16000000L   // 16MHz clock                                                  
  #define INTERRUPT_EXEC_CYCL   90       // Cycles to execute interrupt routines from interrupt.
  #define INTERRUPT_EARLY_BIAS  32       // Cycles to allow of other interrupts.

#elif F_CPU == 8000000L    // 8MHz clock                                                   
  #define INTERRUPT_EXEC_CYCL   45       // Cycles to execute interrupt routines from interrupt.
  #define INTERRUPT_EARLY_BIAS  16       // Cycles to allow of other interrupts.

#else
  #error Unsupported clock speed
#endif


#define PIN_SerialTX		0

#define TRXDDR  DDRD
#define TRXPORT PORTD
#define TRXPIN  PIND

#define GET_RX_PIN( )    ( TRXPIN & ( 1 << PIN_SerialTX ) )

#define ENABLE_TIMER_INTERRUPT( )       ( TIMSK1 |= ( 1<< OCIE1A ) )
#define DISABLE_TIMER_INTERRUPT( )      ( TIMSK1 &= ~( 1<< OCIE1A ) )
#define CLEAR_TIMER_INTERRUPT( )        ( TIFR1 = (1 << OCF1A) )

#define NUMBER_CHANNELS	16
#define	ONE_TO_FOUR			0
#define	FIVE_TO_EIGHT		1

// UART's state.
#define   IDLE				0       // Idle state, both transmit and receive possible.
#define 	PULSING			1				// PPM pulses
#define   RECEIVE     2       // Receiving byte.
#define	  WAITING     3

#define FORCE_INDIRECT(ptr) __asm__ __volatile__ ("" : "=e" (ptr) : "0" (ptr))

uint32_t micros() ;
void setPulseTimes( uint8_t Five2_8 ) ;

//uint16_t FailsafeTimes[NUMBER_CHANNELS] = { 1500, 1500, 1000, 1500, 1500, 1500, 1500, 1500 } ;
uint16_t PulseTimes[NUMBER_CHANNELS] ;

uint8_t *Ports[NUMBER_CHANNELS] = {
	(uint8_t *)&PORTB,
	(uint8_t *)&PORTB,
	(uint8_t *)&PORTB,
	(uint8_t *)&PORTB,
	(uint8_t *)&PORTB,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD,
	(uint8_t *)&PORTD
} ;
uint8_t Bits[NUMBER_CHANNELS] = { 2, 4, 8, 16, 32, 4, 8, 16, 32, 32, 32, 32, 32, 32, 32, 32 } ;

static volatile uint8_t State = IDLE ;     //!< Holds the state of the UART.
static volatile uint8_t Parity ;
static volatile uint8_t SwUartRXData;     //!< Storage for received bits.
static volatile unsigned char SwUartRXBitCount; //!< RX bit counter.
	 
struct t_pulses
{
	uint8_t *port ;
	uint8_t bit ;	
	uint8_t start ;
	uint16_t nextTime ;
} Pulses[8] ;

uint8_t PulsesIndex = 0 ;

uint8_t Sbuffer[28] = {0} ;
uint8_t Sindex = 0 ;
uint16_t Lastrcv = 0 ;
unsigned long Timer = 0 ;

//// Copy the failsafe values into the pulse array
//void enterFailsafe()
//{
//	uint8_t i ;
//	for ( i = 0 ; i < NUMBER_CHANNELS ; i += 1 )
//	{
//		PulseTimes[i] = FailsafeTimes[i] ;
//	}
//}

uint8_t processSBUSframe()
{
	uint8_t inputbitsavailable = 0 ;
	uint8_t i ;
	uint32_t inputbits = 0 ;
	uint8_t *sbus = Sbuffer ;
	if ( Sindex < 23 )
	{
		return 0 ;
	}
	if ( *sbus++ != 0x0F )
	{
		return 0 ;		// Not a valid SBUS frame
	}
	for ( i = 0 ; i < 16 ; i += 1 )
	{
    uint16_t temp ;
		while ( inputbitsavailable < 11 )
		{
			inputbits |= (uint32_t)*sbus++ << inputbitsavailable ;
			inputbitsavailable += 8 ;
		}
    temp = ( (int16_t)( inputbits & 0x7FF ) - 0x3E0 ) * 5 / 8 + 1500 ;
//    cli() ;
		PulseTimes[i] = temp ;
//    sei() ;
		inputbitsavailable -= 11 ;
		inputbits >>= 11 ;
	}
//  if ( timeout )
//  {
//    timeout = 0 ;
//    startTimers() ;
//  }
//  sendSbusFrame() ;
//  Serial.end() ;
//  Serial.begin(100000, SERIAL_8E2) ;
//	Serial.write( buffer, 25 ) ;
//	Serial.flush() ;
//        Serial.end() ;
//  Serial.begin(57600, SERIAL_8N1);
		Sindex = 0 ;
	return 1 ;
}

void setup()	// run once, when the sketch starts
{
//	PulsesIndex = 8 ;	// Past the end, stops the interrupt doing anything
//	enterFailsafe() ;	// Set pulses to failsafe values
//	setPulseTimes( ONE_TO_FOUR ) ;
	PORTB = 0 ;			// Outputs low
	DDRB = 0x3E ;		// 5 outputs
	PORTD = 0 ;     // Outputs low
	DDRD = 0x1C ;		// 3 outputs

	DDRD |= 0x80 ;	// Debug

}

void setPulseTimes( uint8_t Five2_8 )
{
	uint16_t *pulsePtr = PulseTimes ;
	uint16_t times[4] ;				// The 4 pulse lengths to process
	uint8_t i ;
	uint8_t j ;
	uint8_t k ;
	uint16_t m ;
	k = 0 ;								// Offset into Ports and Bits
	if ( Five2_8 )
	{
		pulsePtr += 4 ;			// Move on to second 4 pulses
		k = 4 ;
	}
	for ( i = 0 ; i < 4 ; i += 1 )
	{
		times[i] = pulsePtr[i] ;		// Make local copy of pulses to process
	}
	
	m = times[0] ;								// First pulse
	j = 0 ;
	for ( i = 1 ; i < 4 ; i += 1 )	// Find shortest pulse
	{
		if ( times[i] < m )						// If this one is shorter
		{
			m = times[i] ;							// Note value
			j = i ;											// Note index
		}
	}
	times[j] = 0xFFFF ;							// Make local copy very large
	j += k ;												// Add on offset
	Pulses[4].port = Pulses[0].port = Ports[j] ;	// Set the port for this pulse
	Pulses[4].bit = Pulses[0].bit = Bits[j] ;			// Set the bit for this pulse
	Pulses[0].start = 1 ;													// Mark the start
	Pulses[4].start = 0 ;                         // Mark the end
	uint16_t time = TCNT1 + 1200 ;								// Start the pulses in 75 uS
																								// Gives time for this code to finish
	Pulses[0].nextTime = time + 160 ;							// Next pulse starts in 10 uS
	Pulses[3].nextTime = time + m * 16 ;					// This pulse ends at this time
	
	m = times[0] ;                // First pulse
	j = 0 ;
	for ( i = 1 ; i < 4 ; i += 1 )  // Find next shortest pulse
	{
		if ( times[i] < m )           // If this one is shorter
		{
			m = times[i] ;							// Note value
			j = i ;                     // Note index
		}
	}
	times[j] = 0xFFFF ;             // Make local copy very large
	j += k ;                        // Add on offset
	Pulses[5].port = Pulses[1].port = Ports[j] ;	// Set the port for this pulse
	Pulses[5].bit = Pulses[1].bit = Bits[j] ;     // Set the bit for this pulse
	Pulses[1].start = 1 ;                         // Mark the start
	Pulses[5].start = 0 ;                         // Mark the end
	Pulses[1].nextTime = time + 320 ;             // Next pulse starts in another 10 uS
	Pulses[4].nextTime = time + 160 + m * 16 ;    // This pulse ends at this time
	                                              
	m = times[0] ;                                
	j = 0 ;
	for ( i = 1 ; i < 4 ; i += 1 )
	{
		if ( times[i] < m )
		{
			m = times[i] ;
			j = i ;
		}
	}
	times[j] = 0xFFFF ;
	j += k ;
	Pulses[6].port = Pulses[2].port = Ports[j] ;
	Pulses[6].bit = Pulses[2].bit = Bits[j] ;
	Pulses[2].start = 1 ;
	Pulses[6].start = 0 ;
	Pulses[2].nextTime = time + 480 ;
	Pulses[5].nextTime = time + 320 + m * 16 ;

	m = times[0] ;
	j = 0 ;
	for ( i = 1 ; i < 4 ; i += 1 )
	{
		if ( times[i] < m )
		{
			m = times[i] ;
			j = i ;
		}
	}
	j += k ;
	Pulses[7].port = Pulses[3].port = Ports[j] ;
	Pulses[7].bit = Pulses[3].bit = Bits[j] ;
	Pulses[3].start = 1 ;
	Pulses[7].start = 0 ;
	Pulses[6].nextTime = time + 480 + m * 16 ;
	Pulses[7].nextTime = time + 40000 ;			// Some time well ahead, shouldn't be used
	
	OCR1A = time ;				// Set for first interrupt
}

// ! \brief  External interrupt service routine.  ********************
//  Interrupt on Pin Change to detect change on level on serial signal (= could be a start bit)
//
// The falling edge is the beginning of the start
//  bit will trig this interrupt. The state will
//  be changed to RECEIVE, and the timer interrupt
//  will be set to trig one and a half bit period
//  from the falling edge. At that instant the
//  code should sample the first data bit.
//
//  note  initSoftwareUart( void ) must be called in advance.
//
// This is the pin change interrupt for port D
// This assumes it is the only pin change interrupt
// on this port
ISR(PCINT2_vect)
{
//	if ( TRXPIN & ( 1 << PIN_SerialTX ) )			// Pin is high = start bit (inverted)
#if DEBUGASERIAL
	  PORTC |= 2 ;
#endif
	if ( ( TRXPIN & ( 1 << PIN_SerialTX ) ) == 0 )			// Pin is low = start bit
	{
		PCICR &= ~(1<<PCIE2) ;						// pin change interrupt disabled
//PORTC &= ~2 ;
		State = RECEIVE ;                 // Change state
    DISABLE_TIMER_INTERRUPT() ;       // Disable timer to change its registers.
   	OCR1A = TCNT1 + TICKS2WAITONE_HALF - INTERRUPT_EXEC_CYCL - INTERRUPT_EARLY_BIAS ; // Count one and a half period into the future.
#if DEBUGASERIAL
	  PORTC |= 1 ;
#endif
    SwUartRXBitCount = 0 ;            // Clear received bit counter.
    CLEAR_TIMER_INTERRUPT() ;         // Clear interrupt bits
    ENABLE_TIMER_INTERRUPT() ;        // Enable timer1 interrupt on again
	}
#if DEBUGASERIAL
	  PORTC &= ~2 ;
#endif
}

ISR(TIMER1_COMPA_vect)
{
	if ( State == PULSING )
	{
		uint8_t *port ;
		uint8_t bit ;
		uint16_t index = PulsesIndex ;
		if ( index < 8 )
		{
			port = Pulses[index].port ;				// Output port to modify
			bit = Pulses[index].bit ;					// Output bit to modify
			OCR1A = Pulses[index].nextTime ;	// Time of next action
			if ( Pulses[index].start )				// If starting pulse
			{
				*port |= bit ;									// Set the output
			}
			else
			{
				*port &= ~bit ;									// Else clear it
			}
			PulsesIndex = index + 1 ;					// Next entry
		}
		else
		{
			DISABLE_TIMER_INTERRUPT( ) ;			// We have finished the 4 pulses
			State = IDLE ;
		}
	}
	else
	{
  	switch (State)
		{
  	  case RECEIVE :  // Start bit has been received and we will read bits of data      
  	    OCR1A += TICKS2WAITONE ;                // Count one period after the falling edge is trigged.
        //Receiving, LSB first.
			{
				uint8_t data ;				// Use a temporary local storage
				data = SwUartRXBitCount ;
				if( data < 8 )                         // If 8 bits are not yet read
				{
  	  	  SwUartRXBitCount = data + 1 ;
			  	data = SwUartRXData ;
			    data >>= 1 ;		         // Shift due to receiving LSB first.
	#if DEBUGASERIAL
  	      PORTC &= ~1 ;
	#endif
  	      if( GET_RX_PIN( ) == 1 )
					{
  	        data |= 0x80 ;          // If a logical 1 is read, let the data mirror this.
  	      }
					SwUartRXData = data ;
					Parity += data ;
#if DEBUGASERIAL
  	      PORTC |= 1 ;
#endif
				}
//  	  	else if ( data == 8 )	//Done receiving =  8 bits are in SwUartRXData
//				{
//  	  	  SwUartRXBitCount = data + 1 ;
//					// Could check parity
//					data = Parity ;
//  	      if( GET_RX_PIN( ) == 1 )
//					{
//  	        data ^= 0x80 ;
//  	      }
//    			Sbuffer[Sindex] = SwUartRXData ;
//    			Sindex += 1 ;
//    			Lastrcv = micros() ;
//			    if (Sindex > 27)
//    			{
//      			Sindex = 27 ;
//					}
//				}
				else // Now on stop bit
				{
					DISABLE_TIMER_INTERRUPT() ;		// Stop the timer interrupts.
    			Sbuffer[Sindex] = SwUartRXData ;
    			Lastrcv = TCNT1 ;
    			Sindex += 1 ;
			    if (Sindex > 27)
    			{
      			Sindex = 27 ;
					}
			    State = IDLE ;                // Go back to idle.
					GET_RX_PIN() ;					// Prime pin change
			    PCIFR = (1<<PCIF2) ;					// clear pending interrupt
		  	  PCICR |= (1<<PCIE2) ;					// pin change interrupt enabled
#if DEBUGASERIAL
  	      PORTC &= ~1 ;
#endif
				}
			}
			break ;
  
			case WAITING :
				DISABLE_TIMER_INTERRUPT() ;		// Stop the timer interrupts.
		    State = IDLE ;                // Go back to idle.
				GET_RX_PIN() ;					// Prime pin change
		    PCIFR = (1<<PCIF2) ;					// clear pending interrupt
		    PCICR |= (1<<PCIE2) ;					// pin change interrupt enabled
			break ;

  	// Unknown state.
		  default:        
				State = IDLE;                           // Error, should not occur. Going to a safe state.
  	} // End CASE
	}
} // End of ISR

uint8_t PulsesNeeded ;
uint8_t ppmInState ;
uint8_t ppmInValid ;

void loop()	// run over and over again
{
	uint16_t x ;
	cli() ;
	x = TCNT1 ;
	sei() ;
	if ( ( x - Lastrcv ) > 1000 )
	{
		if ( Sindex > 22 )
		{
			PCICR &= ~(1<<PCIE2) ;						// pin change interrupt disabled
			if ( processSBUSframe() )
			{
				setPulseTimes( ONE_TO_FOUR ) ;	// First 4 pulses
				CLEAR_TIMER_INTERRUPT( ) ;			// Clear flag in case it is set
				PulsesIndex = 0 ;								// Start here
				State = PULSING ;
				ENABLE_TIMER_INTERRUPT( ) ;			// Allow interrupt to run
				Timer = micros() ;
				PulsesNeeded = 1 ;
			}
			else
			{
				DISABLE_TIMER_INTERRUPT( ) ;			// We have finished the 4 pulses
				CLEAR_TIMER_INTERRUPT( ) ;				// Clear flag in case it is set
				State = IDLE ;
				PulsesNeeded = 2 ;                // No longer needed
			}
		}
	}
	
	if ( PulsesNeeded == 1 )		// Second 4 pulses
	{
		if ( ( micros() - Timer ) > 3000 )	// Time for them
		{
			PulsesNeeded = 2 ;                // No longer needed
			setPulseTimes( FIVE_TO_EIGHT ) ;	// Second 4 pulses
			CLEAR_TIMER_INTERRUPT( ) ;				// Clear flag in case it is set
			PulsesIndex = 0 ;									// Start here
			State = PULSING ;
			ENABLE_TIMER_INTERRUPT( ) ;				// Allow interrupt to run
		}
	}
	
	if ( PulsesNeeded == 2 )		// Second 4 pulses
	{
		if ( State == IDLE )
		{
			PulsesNeeded = 0 ;
			GET_RX_PIN() ;					// Prime pin change
		  PCIFR = (1<<PCIF2) ;	// clear pending interrupt
		  PCICR |= (1<<PCIE2) ;	// pin change interrupt enabled
		}
	}
} 


// replacement millis() and micros()
// These work polled, no interrupts
// micros() MUST be called at least once every 4 milliseconds
uint16_t MillisPrecount ;
uint16_t lastTimerValue ;
uint32_t TotalMicros ;
uint32_t TotalMillis ;

uint32_t micros()
{
	uint16_t elapsed ;
	uint8_t millisToAdd ;
	uint8_t oldSREG = SREG ;
	cli() ;
	uint16_t time = TCNT1 ;	// Read timer 1
	SREG = oldSREG ;

	elapsed = time - lastTimerValue ;
	elapsed >>= 4 ;
	
	uint32_t ltime = TotalMicros ;
	ltime += elapsed ;
	cli() ;
	TotalMicros = ltime ;	// Done this way for RPM to work correctly
	lastTimerValue = time ;
	SREG = oldSREG ;	// Still valid from above
	
	elapsed += MillisPrecount;
	millisToAdd = 0 ;
	if ( elapsed  > 3999 )
	{
		millisToAdd = 4 ;
		elapsed -= 4000 ;
	}
	else if ( elapsed  > 2999 )
	{
		millisToAdd = 3 ;		
		elapsed -= 3000 ;
	}
	else if ( elapsed  > 1999 )
	{
		millisToAdd = 2 ;
		elapsed -= 2000 ;
	}
	else if ( elapsed  > 999 )
	{
		millisToAdd = 1 ;
		elapsed -= 1000 ;
	}
	TotalMillis += millisToAdd ;
	MillisPrecount = elapsed ;
	return TotalMicros ;
}

uint32_t millis()
{
	micros() ;
	return TotalMillis ;
}

void initUart()
{
  //PORT
  TRXDDR &= ~( 1 << PIN_SerialTX ) ;       // PIN is input, tri-stated.
  TRXPORT &= ~( 1 << PIN_SerialTX ) ;      // PIN is input, tri-stated.

  // External interrupt
  
#if PIN_SerialTX == 4
  PCMSK2 |= 0x10 ;			// IO4 (PD4) on Arduini mini
#elif PIN_SerialTX == 2
  PCMSK2 |= 0x04 ;  // IO2 (PD2) on Arduini mini
#elif PIN_SerialTX == 0
  PCMSK2 |= 0x01 ;  // IO2 (PD0) on Arduini mini
#else
  #error "This PIN is not supported"
#endif

	GET_RX_PIN() ;					// Prime pin change
  PCIFR = (1<<PCIF2) ;	// clear pending interrupt
  PCICR |= (1<<PCIE2) ;	// pin change interrupt enabled

  // Internal State Variable
  State = IDLE ;

#if DEBUGASERIAL
  DDRC = 0x03 ;		// PC0,1 as o/p debug
  PORTC = 0 ;
#endif
}

void init()
{
  // Timer1
  TCCR1A = 0x00 ;    //Init.
  TCCR1B = 0xC1 ;    // I/p noise cancel, rising edge, Clock/1
	TIFR1 = (1 << OCF1A) ;
	TIMSK1 |= ( 1<< OCIE1A ) ;

	initUart() ;

	sei();

}

void delay(unsigned long ms)
{
	uint16_t start = (uint16_t)micros();
	uint16_t lms = ms ;

	while (lms > 0) {
		if (((uint16_t)micros() - start) >= 1000) {
			lms--;
			start += 1000;
		}
	}
}
 


